#include "cmaes.h"
#include "cmaescb.h"
#include "random.h"

bool notFinish(double f) {
    return false;
}

CMAESC::CMAESC(tCMAESEval eval, int ndim, double minvalue, double maxvalue) : 
    m_dim(ndim), m_eval(eval), m_finish(notFinish) {
       m_mins = m_maxs = NULL;
       m_maxeval = 0;
       setBounds(minvalue, maxvalue);
}

CMAESC::CMAESC(tCMAESEval eval, int ndim) : 
    m_dim(ndim), m_eval(eval), m_finish(notFinish), m_mins(NULL), m_maxs(NULL) {
       m_maxeval = 0;
}


void CMAESC::setBounds(double min, double max) {
    emptyBound();
    m_mins = new Real[m_dim];
    m_maxs = new Real[m_dim];

    fill(m_mins, m_mins+m_dim, min);
    fill(m_maxs, m_maxs+m_dim, max);
}

void CMAESC::setBounds(double *mins, double *maxs) {
    emptyBound();
    m_mins = new Real[m_dim];
    m_maxs = new Real[m_dim];

    copy(mins, mins+m_dim, m_mins);
    copy(maxs, maxs+m_dim, m_maxs);
}



void CMAESC::emptyBound(void) {
    
    if (m_mins != NULL) {
        delete [] m_mins;
	m_mins = NULL;
    }

    if (m_maxs != NULL) {
        delete [] m_maxs;
	m_maxs = NULL;
    }

    m_check_bound = false;
}

CMAESC::~CMAESC(void) {
}



class LocalSearchParamsCMAES : public LocalSearchParams {
public:
    // coordinate wise standard deviation (step size)
    double sigma;
    double maxsigma;
    MyMatrix B;
    MyMatrix BD;
    // covarianza matrix
    MyMatrix C;
    DiagonalMatrix D;
    int m_N;

    // Initialize dynamic (internal) strategy parameters and constants
    ColumnVector pc;
    // evolution paths for C and sigma
    ColumnVector ps;

    LocalSearchParamsCMAES(int N, double sigma) : B(N,N), BD(N,N), C(N,N), D(N), pc(N), ps(N) {
       m_N = N;
       reset(sigma);
    }

    void reset(double sigma) {
        B = eye(m_N);
	D << eye(m_N);
	BD = (B*D);
	C = (BD* BD.t()); // covariance matrix
	pc = 0;
	ps = 0;
	this->sigma = sigma;
	maxsigma = sigma;
    }
};


LocalSearchParamsPtr CMAESC::getInitParams(double initial_sigma) {
   return new LocalSearchParamsCMAES(m_dim, initial_sigma);
}

void CMAESC::setMaxEval(int maxEval) {
    m_maxeval = maxEval;
}


MyReturnMatrix randn(int fil,int col);

/**
 * Esta funcin sirve de interfaz entre la aplicacin de CMAESC desde C y la
 * aplicacin desde MATLAB
 *
 * @param sol Solucin a evaluar, se modific
 * @param m_dim Dimensión a evaluar
 * @param fitness Fitness de la solucin. 
 * @param function Funcin de evaluacin. 
 * @param sigma Desviacin inicial (porcentaje)
 * @param maxeval Mximo nmero de evaluaciones
 * @param m_stop Mnima diferencia para parar (criterio de parada)
 */
int CMAESCapply(double *sol, int m_dim, double &fitness, double *lbounds, double *ubounds, 
	    tCMAESEval eval, void *eval_options,	LocalSearchParamsCMAES *params, double maxsigma,
	    int maxeval, tFinish isFinish, int m_debug);



bool CMAESC::apply(Real *sol, double *pfitness, void *eval_options, LocalSearchParamsPtr params) {
   double fitnessinit = *pfitness;
   double fitness = fitnessinit;
   
   CMAESCapply(sol, m_dim, fitness, m_mins, m_maxs, m_eval, eval_options, (LocalSearchParamsCMAES*) params, -1, m_maxeval, 
		m_finish, 0);
	       
   *pfitness = fitness;
   return (fitness == fitnessinit);
   
}

int debug_at = -1;

/**
 * Esta funcin sirve de interfaz entre la aplicacin de CMAESC desde C y la
 * aplicacin desde MATLAB
 *
 * @param sol Solucin a evaluar, se modific
 * @param m_dim Dimensión a evaluar
 * @param fitness Fitness de la solucin. 
 * @param function Funcin de evaluacin. 
 * @param sigma Desviacin inicial (porcentaje)
 * @param maxeval Mximo nmero de evaluaciones
 * @param m_stop Mnima diferencia para parar (criterio de parada)
 */
int CMAESCapply(double *sol, int m_dim, double &fitness, double *lbounds, double *ubounds, 
	    tCMAESEval eval, void *eval_options,	LocalSearchParamsCMAES *params, double maxsigma,
	    int maxeval, tFinish isFinish, int m_debug) {
   ColumnVector better;
   double fit_better, fit_actual;
   // User defined input parameters (need to be edited)
   int N = m_dim;
   bool converged = false;

   //---------- BEGIN CMAESC ------------------------------------------


   // Inicio en modo avanzado
   CMAESBnd bnd(lbounds, ubounds, N, 1, 1);

   // Active the bound checks if the bounds are valids
   ColumnVector xmean(N);
   xmean << sol;

   better = xmean;
   fit_better = fitness;

   int stopeval = maxeval;         // stop after stopeval number of function evaluations

   // Strategy parameter setting: Selection  
   int lambda=4+(int)floor(3*log(N));  // population size, offspring number
   int mu = (int) floor(lambda/2.0);      // number of parents/points for recombination

   // Indices para escoger los n mejores
   int arindex[lambda];
   int arindex_raw[lambda];

   // muXone array for weighted recombination
   ColumnVector weights(mu);
   int i;

   for (i = 1; i <= mu; i++) {
      weights[i-1] = log(mu+1) - log(i);
   }

   double mueff=pow2(weights.Sum())/pow2(weights).Sum(); //variance-effective size of mu

   if (m_debug) {
      cout <<"xmean: \n" <<xmean <<endl;
      cout <<"C: \n" <<params->C <<endl;
      cout <<"B: \n" <<params->B <<endl;
      cout <<"sigma: " <<params->sigma <<endl;
      cout <<"Lambda: " <<lambda <<endl;
      cout <<"Mu: " <<mu <<endl;
      cout <<"Weights: \n" <<weights <<endl;
      cout <<"mueff " <<mueff <<endl;
   }

   // Strategy parameter setting: Adaptation
   double cc = 4.0/(N+4.0);    // time constant for cumulation for covariance matrix
   double cs = (mueff+2)/(N+mueff+3);   // t-const for cumulation for sigma control
   double mucov = mueff;   // size of mu used for calculating learning rate ccov
   double ccov = (1/mucov) * 2/pow2(N+1.4) + (1-1/mucov)      // learning rate for
     *((2*mueff-1)/(pow2(N+2)+2*mueff));            // covariance matrix

   double damps = 1 + 2*max(0.0, sqrt((mueff-1)/(N+1))-1) + cs; // damping for sigma 
   // usually close to 1
   if (m_debug) {    
      cout <<"cc: " <<cc <<endl;
      cout <<"cs: " <<cs <<endl;
      cout <<"ccov: " <<ccov <<endl;
      cout <<"damps: " <<damps <<endl;
   }

   DiagonalMatrix DiagZeros(N);
   DiagZeros = 0;

   double chiN=sqrt(N)*(1-1/(4.0*N)+1/(21.0*pow2(N)));  // expectation of 
   //   ||N(0,I)|| == norm(randn(N,1))
   //

   weights = weights/weights.Sum();     // normalize recombination weights array

   if (m_debug) {
      cout <<" C : \n" <<params->C <<endl;
      cout <<"chiN : " <<chiN <<endl;
      cout <<"weights : \n" <<weights <<endl;
   }

   // -------------------- Generation Loop --------------------------------

   int counteval = 0;  
   RowVector arfitness(lambda);
   RowVector arfitness_sel(lambda);
   RowVector arfitness_raw(lambda);
   MyMatrix arx(N,lambda);
   MyMatrix arxvalid(N,lambda);
   Matrix arx_bests(N,mu);
   Matrix arz_bests(N,mu);

   // Check if the current step is the last
   bool last = false;
   int countiter = 0;

   while (counteval < stopeval && !isFinish(fit_better)  && !last 
	       // && !converged
	 ) {
      countiter += 1;
      // Generate and evaluate lambda offspring
      MyMatrix arz = randn(N,lambda);  // array of normally distributed mutation vectors

      if (m_debug) {	
	 cout <<"randn:\n" <<arz <<endl;
	 cout <<"xmean:\n" <<xmean <<endl;
	 cout <<"sigma:\n" <<params->sigma <<endl;
	 cout <<"BD:\n" <<params->BD <<endl;
      }

      int k;

      bnd.setParam(lambda, mueff, params->sigma, params->C);

      for (k = 1; k <= lambda && counteval < stopeval; k++) {
	 ColumnVector arx_k = xmean+params->sigma*(params->BD*arz.Column(k)); // add mutation Eq. (1)
	 arx.Column(k) <<arx_k;
	 counteval += 1;
      }

      // Si en vez de lambda hay menos termina
      if (k < lambda) {
	 lambda = min(k,lambda);
	 last = true;
      }

      // Evaluo los nuevos puntos, obtengo dos vectores de fitness
      // el primero contiene el fitness real y el segundo el empleado para
      // la seleccin de la media
      bnd.evalSols(xmean, arx, eval, eval_options, arxvalid, arfitness_raw, arfitness_sel);
      arfitness = arfitness_sel;

      if (counteval == stopeval) {
	 last = true;
      }

      if (m_debug) {	
	 cout <<"arz :\n" <<arz <<endl;
	 cout <<"arx :\n" <<arx <<endl;
	 cout <<"arxvalid :\n" <<arxvalid <<endl;
	 cout <<" arfitness :\n" <<arfitness <<endl;
      }

      range(1, lambda, arindex);
      set_sort_matrix(&arfitness);
      partial_sort(arindex, arindex+mu, arindex+lambda, sort_index_matrix);

      range(1, lambda, arindex_raw);
      set_sort_matrix(&arfitness_raw);
      partial_sort(arindex_raw, arindex_raw+1, arindex_raw+lambda, sort_index_matrix);


      // Actualizo el mejor obtenido hasta el momento
      int newbest = arindex_raw[0];

      if (m_debug) {	
	 cout <<"arfitness[newbest]" <<arfitness_raw[newbest-1] <<endl;
      }

      fit_actual = arfitness_raw[newbest-1];

      if (debug_at != -1 && !m_debug) {
	 m_debug = (counteval == debug_at);
      }

      if (fit_actual < fit_better) {
	 fit_better = fit_actual;
	 better = arxvalid.Column(newbest);
	 params->maxsigma = params->sigma;
      }

      if (last) {
	 copy(better.Store(), better.Store()+N, sol);
	 fitness = fit_better;
	 return counteval;
      }


      // Compruebo la convergencia
      int posconv = (int) ceil(1.0+lambda/4.0);

      if (arfitness[arindex[0]-1] == arfitness[arindex[posconv]-1]) {
	 //	   cout <<"CMAESC: Converged at " <<counteval <<endl;
	 //	   converged = true;
	 params->sigma *= exp(0.2+cs/damps);
      }

      // Update the local search parameters

      //Sort by fitness and compute weighted mean into xmean
      getColumns(arxvalid, arindex, mu, arx_bests);
      getColumns(arz, arindex, mu, arz_bests);

      xmean = arx_bests*weights;   // recombination, new mean value
      MyMatrix zmean = arz_bests*weights;   // == sigma^-1*D^-1*B'*(xmean-xold)
      //  Cumulation: Update evolution paths
      params->ps = (1-cs)*params->ps + sqrt(cs*(2-cs)*mueff) * (params->B * zmean);            // Eq. (4)

      double hsig = norm(params->ps)/pow(sqrt(1-(1-cs)), ((2.0*counteval)/lambda))/chiN < 1.5 + 1/(N+1);

      params->pc = (1-cc)*params->pc + hsig * sqrt(cc*(2-cc)*mueff) * (params->B * params->D * zmean); // Eq. (2)

      if (m_debug) {	
	 cout <<"arx_bests\n" <<arx_bests <<endl;
	 cout <<"arz_bests\n" <<arz_bests <<endl;
	 cout <<"new xmean:\n" <<xmean <<endl;
	 cout <<"zmean:\n" <<zmean <<endl;
	 cout <<"ps:\n" <<params->ps <<endl;
	 cout <<"hsig:\n" <<hsig <<endl;
	 cout <<"pc:\n" <<params->pc <<endl; 
      }

      // Adapt covariance matrix C
      params->C = (1-ccov) * params->C                     // regard old matrix      // Eq. (3)
	+ ccov * (1/mucov) * (params->pc*params->pc.t()     // plus rank one update
		    + (1-hsig) * cc*(2-cc) * params->C) 
	+ ccov * (1-1/mucov) //...           // plus rank mu update 
	* (params->B*params->D*arz_bests) // ...
	*  weights.AsDiagonal() * (params->BD*arz_bests).t();               

      if (m_debug) {
	 cout <<"C:\n" <<params->C <<endl;
      }

      // Adapt step size sigma
      params->sigma = params->sigma * exp((cs/damps)*(norm(params->ps)/chiN - 1));             // Eq. (5)


//       if (ls->checkLocal() && params->sigma > maxsigma) {
	 //	    cout <<"CMAESC Warning: Sigma value too high at " <<counteval <<endl;
//	 last = true;
//      }

      // Update B and D from C
      // This is O(N^3). When strategy internal CPU-time is critical, the
      // next three lines can be executed only every (alpha/ccov/N)-th
      // iteration step, where alpha is e.g. between 0.1 and 10 

      UpperTriangularMatrix triu,triu1;

      triu << params->C; triu1 <<params->C;
      triu1.Inject(DiagZeros);

      if (m_debug) {	
	 cout <<"triu:\n" <<triu <<endl;
	 cout <<"triu1:\n" <<triu1 <<endl;
      }

      params->C = triu+triu1.t(); // Enforce symmetric

      SymmetricMatrix S(N);
      S <<params->C; 

      if (m_debug) {	
	 cout <<"C:\n" <<params->C <<endl;
	 cout <<"S:\n" <<S <<endl;
      }

      EigenValues(S,params->D,params->B);  // eigen decomposition, B==normalized eigenvectors

      if (m_debug) {	
	 cout <<"sigma:\n" <<params->sigma <<endl;
	 cout <<"C:\n" <<params->C <<endl;
	 cout <<"B: \n" <<params->B <<endl;
	 cout <<"D: \n" <<params->D <<endl;
      }

      // Compruebo que explore en toda dimensin
      checkDiag(params->C, params->D);

      params->D << sqrt(params->D);  // D contains standard deviations now

      if (m_debug) {
	 cout <<"Sqrt D: \n" <<params->D <<endl;
	 cout <<counteval <<": " <<arfitness[newbest-1] <<endl;
      }
      // Its value is used again
      params->BD = (params->B*params->D);

      // Compruebo los ejes
      checkAxis(xmean, ccov, cs, damps, countiter, params->sigma, params->C, params->BD);
   }

   if (m_debug) {
      cout <<"xmean: \n" <<xmean <<endl;
      cout <<"better: \n" <<better <<endl;
      cout <<"C: \n" <<params->C <<endl;
      cout <<"B: \n" <<params->B <<endl;
      cout <<"sigma: " <<params->sigma <<endl;
   }

   copy(better.Store(), better.Store()+N, sol);
   fitness = fit_better;

   return counteval;
}

MyReturnMatrix randn(int fil,int col) {
    int total = fil*col;
    double data[total];
    MyMatrix R(fil,col);
    int i;

    for (i = 0; i < total; i++) {
	data[i] = Random::normal(1);
    }

    R <<data;
    R.Release();
    return R;
}

void CMAESC::setFinish(tFinish finish) {
    m_finish = finish;
}

