#ifndef _CMAESC_H 

#define _CMAESC_H 1

#include "newutil.h"

/**
 * Esta funcin permite adaptar la funcin de evaluacin de CMAES
 *
 * (Siempre debe de minimizar). 
 *
 * @param sol Vector de parmetros.
 * @param ndim Tamao del vector
 * @param eval_params Parmetros adiciones de evaluacin
 */
typedef double(*tCMAESEval)(Real *sol, int ndim, void *eval_params); 

/**
 * Esta funcin permite especificar una funcin para parar la bsqueda (como comprobar el ptimo)
 *
 * @param fitness fitness
 * @return si ha terminado la bsqueda
 */
typedef bool(*tFinish)(double);


/**
 * @class LocalSearchParams
 *
 * Esta clase permite guardar los parmetros que determinan la bsqueda. 
 * Almacenar LocalSearchParams permite almacenar los parmetros autoadaptativos
 * en la bsqueda y continuar una bsqueda donde se dej
 */
class LocalSearchParams {

public:
    LocalSearchParams() {
    }
    virtual ~LocalSearchParams() {
    }
    
};

typedef LocalSearchParams* LocalSearchParamsPtr;

/**
 * @class CMAESC 
 * Permite aplicar la BL CMAESC (contenida en libcmaes.so) implementada en 
 * Matlab, por lo que posee unas cuantas caractersticas especiales. 
 *
 * CMAESC es un EDA que ha obtenido los mejores resultados en codificacin real
 * en el CEC2005. Presenta los mejores resultados en Unimodales, sus resultados 
 * en Multimodales, por contra, son mejorables. Espero que introducirlo en 
 * este framework posibilitar buena sinergia entre el esquema de BG y el esquema de BL del CMAESC. 
 * 
 */
class CMAESC {
public:
    /**
     * Constructor de la clase
     *
     * @param eval Criterio de optimizacin
     * @param ndim Tamao de cada solucin
     * @param minvalue Mnimo valor a explorar
     * @param maxvalue Mximo valor a explorar
     */
    CMAESC(tCMAESEval eval, int ndim, double minvalue, double maxvalue);

    CMAESC(tCMAESEval eval, int ndim);


    /**
     * Obtengo los parmetros iniciales
     * @param double init_sigma Parmetro sigma inicial
     */
    LocalSearchParamsPtr getInitParams(double initial_sigma); 

    /**
     * Permite indicar el nmero de evaluaciones
     */
    void setMaxEval(int maxeval);

    /**
     * Aplico la BL sobre una solucin
     *
     * @param sol vector solucin (se modificar)
     * @param fitness Fitness (valor fitness de la mejor solucin encontrada)
     * @param eval_options Parmetros opcionales
     * @param params Parmetros de la BL (se modificar)
     */
    bool apply(Real *sol, double *fitness, void *eval_options, LocalSearchParamsPtr params);

    /**
     * Este mtodo permite especificar los valores mximos y mnimos
     *
     * @param min Mnimo valor de la bsqueda
     * @param max Mximo valor de la bsqueda
     */
    void setBounds(double min, double max);

    /**
     * Este mtodo permite especificar los valores mximos y mnimos
     *
     * @param min Mnimo valor de la bsqueda
     * @param max Mximo valor de la bsqueda
     */
   
    void setBounds(double *mins, double *maxs);

    /**
     * Este mtodo elimina la comprobacin de frontera
     */
    void emptyBound(void); 

    virtual ~CMAESC(void);

    /**
     * Esta funcin permite especificar una funcin para parar la bsqueda
     *
     */
    void setFinish(tFinish finish);

    
protected:
    int m_dim;
    int m_maxeval;
    tCMAESEval m_eval;
    tFinish m_finish;
    double *m_mins;
    double *m_maxs;
    bool m_check_bound;
    
};


#endif
